@@ -301,8 +301,10 @@ _cairo_gl_pattern_texture_setup (cairo_gl_operand_t *operand,
        so we need to release this device while we paint it to the image. */
     if (src_is_gl_surface) {
 	status = _cairo_gl_context_release (ctx, status);
-	if (unlikely (status))
+	if (unlikely (status)) {
+	    _cairo_surface_unmap_image (&surface->base, image);
 	    goto fail;
+	}
     }
 
     status = _cairo_surface_offset_paint (&image->base, extents->x, extents->y,
@@ -310,8 +312,10 @@ _cairo_gl_pattern_texture_setup (cairo_gl_operand_t *operand,
 
     if (src_is_gl_surface) {
 	status = _cairo_gl_context_acquire (dst->base.device, &ctx);
-	if (unlikely (status))
+	if (unlikely (status)) {
+	    _cairo_surface_unmap_image (&surface->base, image);
 	    goto fail;
+	}
     }
 
     status = _cairo_surface_unmap_image (&surface->base, image);
@@ -654,7 +658,8 @@ _cairo_gl_operand_bind_to_shader (cairo_gl_context_t *ctx,
 	 * with CAIRO_EXTEND_NONE). When bilinear filtering is enabled,
 	 * these shaders need the texture dimensions for their calculations.
 	 */
-	if (ctx->gl_flavor == CAIRO_GL_FLAVOR_ES &&
+	if ((ctx->gl_flavor == CAIRO_GL_FLAVOR_ES3 ||
+	     ctx->gl_flavor == CAIRO_GL_FLAVOR_ES2) &&
 	    _cairo_gl_operand_get_extend (operand) == CAIRO_EXTEND_NONE &&
 	    _cairo_gl_operand_get_gl_filter (operand) == GL_LINEAR)
 	{
