@@ -158,6 +158,14 @@ png_simple_warning_callback (png_structp png,
      */
 }
 
+static int
+png_setjmp (png_struct *png)
+{
+#ifdef PNG_SETJMP_SUPPORTED
+    return setjmp (png_jmpbuf (png));
+#endif
+    return 0;
+}
 
 /* Starting with libpng-1.2.30, we must explicitly specify an output_flush_fn.
  * Otherwise, we will segfault if we are writing to a stream. */
@@ -229,10 +237,8 @@ write_png (cairo_surface_t	*surface,
 	goto BAIL4;
     }
 
-#ifdef PNG_SETJMP_SUPPORTED
-    if (setjmp (png_jmpbuf (png)))
+    if (png_setjmp (png))
 	goto BAIL4;
-#endif
 
     png_set_write_fn (png, closure, write_func, png_simple_output_flush_fn);
 
@@ -339,7 +345,8 @@ stdio_write_func (png_structp png, png_bytep data, png_size_t size)
 /**
  * cairo_surface_write_to_png:
  * @surface: a #cairo_surface_t with pixel contents
- * @filename: the name of a file to write to
+ * @filename: the name of a file to write to; on Windows this filename
+ *   is encoded in UTF-8.
  *
  * Writes the contents of @surface to a new file @filename as a PNG
  * image.
@@ -367,7 +374,11 @@ cairo_surface_write_to_png (cairo_surface_t	*surface,
     if (surface->finished)
 	return _cairo_error (CAIRO_STATUS_SURFACE_FINISHED);
 
-    fp = fopen (filename, "wb");
+    status = _cairo_fopen (filename, "wb", &fp);
+
+    if (status != CAIRO_STATUS_SUCCESS)
+	return _cairo_error (status);
+
     if (fp == NULL) {
 	switch (errno) {
 	case ENOMEM:
@@ -573,12 +584,11 @@ read_png (struct png_read_closure_t *png_closure)
     png_set_read_fn (png, png_closure, stream_read_func);
 
     status = CAIRO_STATUS_SUCCESS;
-#ifdef PNG_SETJMP_SUPPORTED
-    if (setjmp (png_jmpbuf (png))) {
+
+    if (png_setjmp (png)) {
 	surface = _cairo_surface_create_in_error (status);
 	goto BAIL;
     }
-#endif
 
     png_read_info (png, info);
 
@@ -673,7 +683,7 @@ read_png (struct png_read_closure_t *png_closure)
     }
 
     for (i = 0; i < png_height; i++)
-        row_pointers[i] = &data[i * stride];
+        row_pointers[i] = &data[i * (ptrdiff_t)stride];
 
     png_read_image (png, row_pointers);
     png_read_end (png, info);
@@ -733,7 +743,8 @@ read_png (struct png_read_closure_t *png_closure)
 
 /**
  * cairo_image_surface_create_from_png:
- * @filename: name of PNG file to load
+ * @filename: name of PNG file to load. On Windows this filename
+ *   is encoded in UTF-8.
  *
  * Creates a new image surface and initializes the contents to the
  * given PNG file.
@@ -759,10 +770,14 @@ cairo_image_surface_create_from_png (const char *filename)
 {
     struct png_read_closure_t png_closure;
     cairo_surface_t *surface;
+    cairo_status_t status;
+
+    status = _cairo_fopen (filename, "rb", (FILE **) &png_closure.closure);
+
+    if (status != CAIRO_STATUS_SUCCESS)
+	return _cairo_surface_create_in_error (status);
 
-    png_closure.closure = fopen (filename, "rb");
     if (png_closure.closure == NULL) {
-	cairo_status_t status;
 	switch (errno) {
 	case ENOMEM:
 	    status = _cairo_error (CAIRO_STATUS_NO_MEMORY);
