@@ -39,7 +39,7 @@
  *	Carl Worth <cworth@cworth.org>
  */
 
-#define _BSD_SOURCE /* for snprintf() */
+#define _DEFAULT_SOURCE /* for snprintf() */
 #include "cairoint.h"
 
 #include "cairo-svg.h"
@@ -94,6 +94,7 @@ static const char *_cairo_svg_supported_mime_types[] =
 {
     CAIRO_MIME_TYPE_JPEG,
     CAIRO_MIME_TYPE_PNG,
+    CAIRO_MIME_TYPE_UNIQUE_ID,
     CAIRO_MIME_TYPE_URI,
     NULL
 };
@@ -121,6 +122,20 @@ static const char * _cairo_svg_internal_version_strings[CAIRO_SVG_VERSION_LAST]
     "1.2"
 };
 
+static const char * _cairo_svg_unit_strings[] =
+{
+    [CAIRO_SVG_UNIT_USER] = "",
+    [CAIRO_SVG_UNIT_EM] = "em",
+    [CAIRO_SVG_UNIT_EX] = "ex",
+    [CAIRO_SVG_UNIT_PX] = "px",
+    [CAIRO_SVG_UNIT_IN] = "in",
+    [CAIRO_SVG_UNIT_CM] = "cm",
+    [CAIRO_SVG_UNIT_MM] = "mm",
+    [CAIRO_SVG_UNIT_PT] = "pt",
+    [CAIRO_SVG_UNIT_PC] = "pc",
+    [CAIRO_SVG_UNIT_PERCENT] = "%"
+};
+
 struct cairo_svg_page {
     unsigned int surface_id;
     unsigned int clip_level;
@@ -135,6 +150,7 @@ struct cairo_svg_document {
 
     double width;
     double height;
+    cairo_svg_unit_t unit;
 
     cairo_output_stream_t *xml_node_defs;
     cairo_output_stream_t *xml_node_glyphs;
@@ -176,7 +192,9 @@ static cairo_surface_t *
 _cairo_svg_surface_create_for_document (cairo_svg_document_t	*document,
 					cairo_content_t		 content,
 					double			 width,
-					double			 height);
+					double			 height,
+					cairo_bool_t             bounded);
+
 static cairo_surface_t *
 _cairo_svg_surface_create_for_stream_internal (cairo_output_stream_t	*stream,
 					       double			 width,
@@ -252,6 +270,9 @@ cairo_svg_surface_create_for_stream (cairo_write_func_t		 write_func,
  * or %CAIRO_MIME_TYPE_PNG is specified, the corresponding data is
  * Base64-encoded and emitted.
  *
+ * If %CAIRO_MIME_TYPE_UNIQUE_ID is present, all surfaces with the same
+ * unique identifier will only be embedded once.
+ *
  * Return value: a pointer to the newly created surface. The caller
  * owns the surface and should call cairo_surface_destroy() when done
  * with it.
@@ -399,6 +420,164 @@ cairo_svg_version_to_string (cairo_svg_version_t version)
     return _cairo_svg_version_strings[version];
 }
 
+/**
+ * cairo_svg_surface_set_document_unit:
+ * @surface: a SVG #cairo_surface_t
+ * @unit: SVG unit
+ *
+ * Use the specified unit for the width and height of the generated SVG file.
+ * See #cairo_svg_unit_t for a list of available unit values that can be used
+ * here.
+ *
+ * This function can be called at any time before generating the SVG file.
+ *
+ * However to minimize the risk of ambiguities it's recommended to call it
+ * before any drawing operations have been performed on the given surface, to
+ * make it clearer what the unit used in the drawing operations is.
+ *
+ * The simplest way to do this is to call this function immediately after
+ * creating the SVG surface.
+ *
+ * Note if this function is never called, the default unit for SVG documents
+ * generated by cairo will be "pt". This is for historical reasons.
+ *
+ * Since: 1.16
+ **/
+void
+cairo_svg_surface_set_document_unit (cairo_surface_t	*abstract_surface,
+				     cairo_svg_unit_t	 unit)
+{
+    cairo_svg_surface_t *surface = NULL; /* hide compiler warning */
+
+    if (! _extract_svg_surface (abstract_surface, &surface))
+	return;
+
+    if (unit <= CAIRO_SVG_UNIT_PERCENT)
+	surface->document->unit = unit;
+}
+
+/**
+ * cairo_svg_surface_get_document_unit:
+ * @surface: a SVG #cairo_surface_t
+ *
+ * Get the unit of the SVG surface.
+ *
+ * If the surface passed as an argument is not a SVG surface, the function
+ * sets the error status to CAIRO_STATUS_SURFACE_TYPE_MISMATCH and returns
+ * CAIRO_SVG_UNIT_USER.
+ *
+ * Return value: the SVG unit of the SVG surface.
+ *
+ * Since: 1.16
+ **/
+cairo_svg_unit_t
+cairo_svg_surface_get_document_unit (cairo_surface_t	*abstract_surface)
+{
+    cairo_svg_surface_t *surface = NULL; /* hide compiler warning */
+
+    if (! _extract_svg_surface (abstract_surface, &surface)) {
+	_cairo_error_throw (CAIRO_STATUS_SURFACE_TYPE_MISMATCH);
+	return CAIRO_SVG_UNIT_USER;
+    }
+
+    return surface->document->unit;
+}
+
+static void
+_cairo_svg_source_surface_init_key (cairo_svg_source_surface_t *key)
+{
+    if (key->unique_id && key->unique_id_length > 0) {
+	key->base.hash = _cairo_hash_bytes (_CAIRO_HASH_INIT_VALUE,
+					    key->unique_id, key->unique_id_length);
+    } else {
+	key->base.hash = key->id;
+    }
+}
+
+static cairo_bool_t
+_cairo_svg_source_surface_equal (const void *key_a, const void *key_b)
+{
+    const cairo_svg_source_surface_t *a = key_a;
+    const cairo_svg_source_surface_t *b = key_b;
+
+    if (a->unique_id && b->unique_id && a->unique_id_length == b->unique_id_length)
+	return (memcmp (a->unique_id, b->unique_id, a->unique_id_length) == 0);
+
+    return (a->id == b->id);
+}
+
+static void
+_cairo_svg_source_surface_pluck (void *entry, void *closure)
+{
+    cairo_svg_source_surface_t *surface_entry = entry;
+    cairo_hash_table_t *patterns = closure;
+
+    _cairo_hash_table_remove (patterns, &surface_entry->base);
+    free (surface_entry->unique_id);
+    free (surface_entry);
+}
+
+static cairo_status_t
+_cairo_svg_surface_add_source_surface (cairo_svg_surface_t  *surface,
+				       cairo_surface_t	    *source_surface,
+				       int                  *source_id,
+				       cairo_bool_t         *is_new)
+{
+    cairo_svg_source_surface_t source_key;
+    cairo_svg_source_surface_t *source_entry;
+    unsigned char *unique_id = NULL;
+    unsigned long unique_id_length = 0;
+    cairo_status_t status;
+
+    source_key.id  = source_surface->unique_id;
+    cairo_surface_get_mime_data (source_surface, CAIRO_MIME_TYPE_UNIQUE_ID,
+				 (const unsigned char **) &source_key.unique_id,
+				 &source_key.unique_id_length);
+    _cairo_svg_source_surface_init_key (&source_key);
+    source_entry = _cairo_hash_table_lookup (surface->source_surfaces, &source_key.base);
+    if (source_entry) {
+	*source_id = source_entry->id;
+	*is_new = FALSE;
+	return CAIRO_STATUS_SUCCESS;
+    }
+
+    if (source_key.unique_id && source_key.unique_id_length > 0) {
+	unique_id = _cairo_malloc (source_key.unique_id_length);
+	if (unique_id == NULL) {
+	    return _cairo_error (CAIRO_STATUS_NO_MEMORY);
+	}
+
+	unique_id_length = source_key.unique_id_length;
+	memcpy (unique_id, source_key.unique_id, unique_id_length);
+    } else {
+	unique_id = NULL;
+	unique_id_length = 0;
+    }
+
+    source_entry = malloc (sizeof (cairo_svg_source_surface_t));
+    if (source_entry == NULL) {
+	status = _cairo_error (CAIRO_STATUS_NO_MEMORY);
+	goto fail;
+    }
+
+    source_entry->id = source_key.id;
+    source_entry->unique_id_length = unique_id_length;
+    source_entry->unique_id = unique_id;
+    _cairo_svg_source_surface_init_key (source_entry);
+    status = _cairo_hash_table_insert (surface->source_surfaces, &source_entry->base);
+    if (unlikely(status))
+	goto fail;
+
+    *is_new = TRUE;
+    *source_id = source_entry->id;
+    return CAIRO_STATUS_SUCCESS;
+
+  fail:
+    free (unique_id);
+    free (source_entry);
+    return status;
+}
+
 static cairo_bool_t
 _cliprect_covers_surface (cairo_svg_surface_t *surface,
 			  cairo_path_fixed_t *path)
@@ -470,7 +649,8 @@ static cairo_surface_t *
 _cairo_svg_surface_create_for_document (cairo_svg_document_t	*document,
 					cairo_content_t		 content,
 					double			 width,
-					double			 height)
+					double			 height,
+					cairo_bool_t             bounded)
 {
     cairo_svg_surface_t *surface;
     cairo_surface_t *paginated;
@@ -488,6 +668,7 @@ _cairo_svg_surface_create_for_document (cairo_svg_document_t	*document,
 
     surface->width = width;
     surface->height = height;
+    surface->surface_bounded = bounded;
 
     surface->document = _cairo_svg_document_reference (document);
 
@@ -520,6 +701,12 @@ _cairo_svg_surface_create_for_document (cairo_svg_document_t	*document,
     surface->force_fallbacks = FALSE;
     surface->content = content;
 
+    surface->source_surfaces = _cairo_hash_table_create (_cairo_svg_source_surface_equal);
+    if (unlikely (surface->source_surfaces == NULL)) {
+	status = _cairo_error (CAIRO_STATUS_NO_MEMORY);
+	goto CLEANUP;
+    }
+
     paginated = _cairo_paginated_surface_create (&surface->base,
 					         surface->content,
 						 &cairo_svg_surface_paginated_backend);
@@ -561,7 +748,7 @@ _cairo_svg_surface_create_for_stream_internal (cairo_output_stream_t	*stream,
     }
 
     surface = _cairo_svg_surface_create_for_document (document, CAIRO_CONTENT_COLOR_ALPHA,
-						      width, height);
+						      width, height, TRUE);
     if (surface->status) {
 	status = _cairo_svg_document_destroy (document);
 	return surface;
@@ -1011,6 +1198,11 @@ _cairo_svg_surface_finish (void *abstract_surface)
 
     _cairo_surface_clipper_reset (&surface->clipper);
 
+    _cairo_hash_table_foreach (surface->source_surfaces,
+			       _cairo_svg_source_surface_pluck,
+			       surface->source_surfaces);
+    _cairo_hash_table_destroy (surface->source_surfaces);
+
     status2 = _cairo_svg_document_destroy (document);
     if (status == CAIRO_STATUS_SUCCESS)
 	status = status2;
@@ -1282,7 +1474,8 @@ _cairo_svg_surface_emit_attr_value (cairo_output_stream_t *stream,
 
 static cairo_status_t
 _cairo_svg_surface_emit_surface (cairo_svg_document_t *document,
-				 cairo_surface_t *surface)
+				 cairo_surface_t *surface,
+				 int source_id)
 {
     cairo_rectangle_int_t extents;
     cairo_bool_t is_bounded;
@@ -1290,18 +1483,12 @@ _cairo_svg_surface_emit_surface (cairo_svg_document_t *document,
     const unsigned char *uri;
     unsigned long uri_len;
 
-    if (_cairo_user_data_array_get_data (&surface->user_data,
-					 (cairo_user_data_key_t *) document))
-    {
-	return CAIRO_STATUS_SUCCESS;
-    }
-
     is_bounded = _cairo_surface_get_extents (surface, &extents);
     assert (is_bounded);
 
     _cairo_output_stream_printf (document->xml_node_defs,
 				 "<image id=\"image%d\" width=\"%d\" height=\"%d\"",
-				 surface->unique_id,
+				 source_id,
 				 extents.width, extents.height);
 
     _cairo_output_stream_printf (document->xml_node_defs, " xlink:href=\"");
@@ -1320,10 +1507,7 @@ _cairo_svg_surface_emit_surface (cairo_svg_document_t *document,
 
     _cairo_output_stream_printf (document->xml_node_defs, "\"/>\n");
 
-    /* and tag it */
-    return _cairo_user_data_array_set_data (&surface->user_data,
-					    (cairo_user_data_key_t *) document,
-					    document, NULL);
+    return CAIRO_STATUS_SUCCESS;
 }
 
 static cairo_status_t
@@ -1337,17 +1521,29 @@ _cairo_svg_surface_emit_composite_surface_pattern (cairo_output_stream_t   *outp
 {
     cairo_status_t status;
     cairo_matrix_t p2u;
+    int source_id;
+    cairo_bool_t is_new;
 
     p2u = pattern->base.matrix;
     status = cairo_matrix_invert (&p2u);
     /* cairo_pattern_set_matrix ensures the matrix is invertible */
     assert (status == CAIRO_STATUS_SUCCESS);
 
-    status = _cairo_svg_surface_emit_surface (svg_surface->document,
-					      pattern->surface);
+    status = _cairo_svg_surface_add_source_surface (svg_surface,
+						    pattern->surface,
+						    &source_id,
+						    &is_new);
     if (unlikely (status))
 	return status;
 
+    if (is_new) {
+	status = _cairo_svg_surface_emit_surface (svg_surface->document,
+						  pattern->surface,
+						  source_id);
+	if (unlikely (status))
+	    return status;
+    }
+
     if (pattern_id != invalid_pattern_id) {
 	cairo_rectangle_int_t extents;
 	cairo_bool_t is_bounded;
@@ -1369,7 +1565,7 @@ _cairo_svg_surface_emit_composite_surface_pattern (cairo_output_stream_t   *outp
 
     _cairo_output_stream_printf (output,
 				 "<use xlink:href=\"#image%d\"",
-				 pattern->surface->unique_id);
+				 source_id);
     if (extra_attributes)
 	_cairo_output_stream_printf (output, " %s", extra_attributes);
 
@@ -1390,25 +1586,23 @@ _cairo_svg_surface_emit_composite_surface_pattern (cairo_output_stream_t   *outp
 
 static cairo_status_t
 _cairo_svg_surface_emit_recording_surface (cairo_svg_document_t      *document,
-					   cairo_recording_surface_t *source)
+					   cairo_recording_surface_t *source,
+					   int                        source_id)
 {
     cairo_status_t status;
     cairo_surface_t *paginated_surface;
     cairo_svg_surface_t *svg_surface;
     cairo_array_t *page_set;
-
+    cairo_rectangle_int_t extents;
+    cairo_bool_t bounded;
     cairo_output_stream_t *contents;
 
-    if (_cairo_user_data_array_get_data (&source->base.user_data,
-					 (cairo_user_data_key_t *) document))
-    {
-	return CAIRO_STATUS_SUCCESS;
-    }
-
+    bounded = _cairo_surface_get_extents (&source->base, &extents);
     paginated_surface = _cairo_svg_surface_create_for_document (document,
 								source->base.content,
-								source->extents_pixels.width,
-								source->extents_pixels.height);
+								extents.width,
+								extents.height,
+								bounded);
     if (unlikely (paginated_surface->status))
 	return paginated_surface->status;
 
@@ -1436,13 +1630,17 @@ _cairo_svg_surface_emit_recording_surface (cairo_svg_document_t      *document,
 
     if (! svg_surface->is_base_clip_emitted) {
 	svg_surface->is_base_clip_emitted = TRUE;
-	_cairo_output_stream_printf (document->xml_node_defs,
-				     "<clipPath id=\"clip%d\">\n"
-				     "  <rect width=\"%f\" height=\"%f\"/>\n"
-				     "</clipPath>\n",
-				     svg_surface->base_clip,
-				     svg_surface->width,
-				     svg_surface->height);
+	if (_cairo_surface_get_extents (&svg_surface->base, &extents)) {
+	    _cairo_output_stream_printf (document->xml_node_defs,
+					 "<clipPath id=\"clip%d\">\n"
+					 "  <rect x=\"%d\" y=\"%d\" width=\"%d\" height=\"%d\"/>\n"
+					 "</clipPath>\n",
+					 svg_surface->base_clip,
+					 extents.x,
+					 extents.y,
+					 extents.width,
+					 extents.height);
+	}
     }
 
     if (source->base.content == CAIRO_CONTENT_ALPHA) {
@@ -1451,13 +1649,13 @@ _cairo_svg_surface_emit_recording_surface (cairo_svg_document_t      *document,
 				     "<g id=\"surface%d\" "
 				     "clip-path=\"url(#clip%d)\" "
 				     "filter=\"url(#alpha)\">\n",
-				     source->base.unique_id,
+				     source_id,
 				     svg_surface->base_clip);
     } else {
 	_cairo_output_stream_printf (document->xml_node_defs,
 				     "<g id=\"surface%d\" "
 				     "clip-path=\"url(#clip%d)\">\n",
-				     source->base.unique_id,
+				     source_id,
 				     svg_surface->base_clip);
     }
 
@@ -1483,13 +1681,7 @@ _cairo_svg_surface_emit_recording_surface (cairo_svg_document_t      *document,
     status = cairo_surface_status (paginated_surface);
     cairo_surface_destroy (paginated_surface);
 
-    if (unlikely (status))
-	return status;
-
-    /* and tag it */
-    return _cairo_user_data_array_set_data (&source->base.user_data,
-					    (cairo_user_data_key_t *) document,
-					    document, NULL);
+    return status;
 }
 
 static cairo_recording_surface_t *
@@ -1516,17 +1708,28 @@ _cairo_svg_surface_emit_composite_recording_pattern (cairo_output_stream_t	*outp
     cairo_recording_surface_t *recording_surface;
     cairo_matrix_t p2u;
     cairo_status_t status;
+    int source_id;
+    cairo_bool_t is_new;
 
     p2u = pattern->base.matrix;
     status = cairo_matrix_invert (&p2u);
     /* cairo_pattern_set_matrix ensures the matrix is invertible */
     assert (status == CAIRO_STATUS_SUCCESS);
 
-    recording_surface = to_recording_surface (pattern);
-    status = _cairo_svg_surface_emit_recording_surface (document, recording_surface);
+    status = _cairo_svg_surface_add_source_surface (surface,
+						    pattern->surface,
+						    &source_id,
+						    &is_new);
     if (unlikely (status))
 	return status;
 
+    recording_surface = to_recording_surface (pattern);
+    if (is_new) {
+	status = _cairo_svg_surface_emit_recording_surface (document, recording_surface, source_id);
+	if (unlikely (status))
+	    return status;
+    }
+
     if (pattern_id != invalid_pattern_id) {
 	_cairo_output_stream_printf (output,
 				     "<pattern id=\"pattern%d\" "
@@ -1541,7 +1744,7 @@ _cairo_svg_surface_emit_composite_recording_pattern (cairo_output_stream_t	*outp
 
     _cairo_output_stream_printf (output,
 				 "<use xlink:href=\"#surface%d\"",
-				 recording_surface->base.unique_id);
+				 source_id);
 
     if (pattern_id == invalid_pattern_id) {
 	_cairo_svg_surface_emit_operator (output, surface, op);
@@ -2258,7 +2461,7 @@ _cairo_svg_surface_get_extents (void		        *abstract_surface,
     rectangle->width  = ceil (surface->width);
     rectangle->height = ceil (surface->height);
 
-    return TRUE;
+    return surface->surface_bounded;
 }
 
 static cairo_status_t
@@ -2676,6 +2879,7 @@ _cairo_svg_document_create (cairo_output_stream_t	 *output_stream,
     document->finished = FALSE;
     document->width = width;
     document->height = height;
+    document->unit = CAIRO_SVG_UNIT_PT;
 
     document->linear_pattern_id = 0;
     document->radial_pattern_id = 0;
@@ -2778,9 +2982,10 @@ _cairo_svg_document_finish (cairo_svg_document_t *document)
 				 "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
 				 "<svg xmlns=\"http://www.w3.org/2000/svg\" "
 				 "xmlns:xlink=\"http://www.w3.org/1999/xlink\" "
-				 "width=\"%fpt\" height=\"%fpt\" "
+				 "width=\"%f%s\" height=\"%f%s\" "
 				 "viewBox=\"0 0 %f %f\" version=\"%s\">\n",
-				 document->width, document->height,
+				 document->width, _cairo_svg_unit_strings [document->unit],
+				 document->height, _cairo_svg_unit_strings [document->unit],
 				 document->width, document->height,
 				 _cairo_svg_internal_version_strings [document->svg_version]);
 
